import os


AUTO_LOADED_PROFILES = list(os.path.join(*parts) for parts in (
    ('.landscape.yml',),
    ('.landscape.yaml',),
    ('landscape.yml',),
    ('landscape.yaml',),
    ('.prospector.yaml',),
    ('.prospector.yml',),
    ('prospector.yaml',),
    ('prospector.yml',),
    ('prospector', '.prospector.yaml'),
    ('prospector', '.prospector.yml'),
    ('prospector', 'prospector.yaml'),
    ('prospector', 'prospector.yml'),
    ('.prospector', '.prospector.yaml'),
    ('.prospector', '.prospector.yml'),
    ('.prospector', 'prospector.yaml'),
    ('.prospector', 'prospector.yml'),
))



from __future__ import absolute_import

from pylint.__pkginfo__ import numversion as PYLINT_VERSION
if PYLINT_VERSION >= (1, 5):
    from pylint.config import OptionsManagerMixIn
    from pylint.utils import _splitstrip
else:
    from logilab.common.configuration import OptionsManagerMixIn
from pylint.lint import PyLinter


class ProspectorLinter(PyLinter):  # pylint: disable=too-many-ancestors,too-many-public-methods

    def __init__(self, found_files, *args, **kwargs):
        self._files = found_files

        # set up the standard PyLint linter
        PyLinter.__init__(self, *args, **kwargs)

    def config_from_file(self, config_file=None):
        """Will return `True` if plugins have been loaded. For pylint>=1.5. Else `False`."""
        if PYLINT_VERSION >= (1, 5):
            if PYLINT_VERSION >= (2, 0):
                self.read_config_file(config_file)
            else:
                self.read_config_file(config_file, quiet=True)
            if self.cfgfile_parser.has_option('MASTER', 'load-plugins'):
                # pylint: disable=protected-access
                plugins = _splitstrip(self.cfgfile_parser.get('MASTER', 'load-plugins'))
                self.load_plugin_modules(plugins)
            self.load_config_file()
            return True

        self.load_file_configuration(config_file)
        return False

    def reset_options(self):
        # for example, we want to re-initialise the OptionsManagerMixin
        # to supress the config error warning
        # pylint: disable=non-parent-init-called
        if PYLINT_VERSION >= (2, 0):
            OptionsManagerMixIn.__init__(self, usage=PyLinter.__doc__)
        else:
            OptionsManagerMixIn.__init__(self, usage=PyLinter.__doc__, quiet=True)

    def expand_files(self, modules):
        expanded = PyLinter.expand_files(self, modules)
        filtered = []
        for module in expanded:
            if self._files.check_module(module['path']):
                filtered.append(module)
        return filtered



from os import pathsep

from setoptconf.datatype import Choice


class OutputChoice(Choice):
    def sanitize(self, value):
        splitted_value = value.split(pathsep)
        output_format, output_targets = splitted_value[0], splitted_value[1:]
        validated_format =  super(OutputChoice, self).sanitize(output_format)
        return (output_format, output_targets)


from collections import defaultdict

from prospector.formatters.text import TextFormatter


__all__ = (
    'GroupedFormatter',
)


class GroupedFormatter(TextFormatter):
    def render_messages(self):
        output = [
            'Messages',
            '========',
            '',
        ]

        # pylint: disable=unnecessary-lambda
        groups = defaultdict(lambda: defaultdict(list))

        for message in self.messages:
            groups[message.location.path][message.location.line].append(message)

        for filename in sorted(groups.keys()):
            output.append(filename)

            for line in sorted(groups[filename].keys(), key=lambda x: 0 if x is None else int(x)):
                output.append('  Line: %s' % line)

                for message in groups[filename][line]:
                    output.append(
                        '    %s: %s / %s%s' % (
                            message.source,
                            message.code,
                            message.message,
                            (' (col %s)' % message.location.character) if message.location.character else '',
                        )
                    )

            output.append('')

        return '\n'.join(output)



# -*- coding: utf-8 -*-
from itertools import islice
from mypy import api

from prospector.message import Location, Message
from prospector.tools import ToolBase


__all__ = (
    'MypyTool',
)


MYPY_OPTIONS = [
    'allow',
    'check',
    'disallow',
    'no-check',
    'no-warn',
    'warn'
]


class MypyTool(ToolBase):
    def __init__(self, *args, **kwargs):
        super(MypyTool, self).__init__(*args, **kwargs)
        self.checker = api
        self.options = ['--show-column-numbers']

    def configure(self, prospector_config, _):
        options = prospector_config.tool_options('mypy')

        follow_imports = options.get('follow-imports', 'normal')
        ignore_missing_imports = options.get('ignore-missing-imports', False)
        implict_optional = options.get('implict-optional', False)
        platform = options.get('platform', None)
        python_2_mode = options.get('python-2-mode', False)
        python_version = options.get('python-version', None)
        strict_optional = options.get('strict-optional', False)

        self.options.append('--follow-imports=%s' % follow_imports)

        if ignore_missing_imports:
            self.options.append('--ignore-missing-imports')

        if implict_optional:
            self.options.append('--implict-optional')

        if platform:
            self.options.append('--platform %s' % platform)

        if python_2_mode:
            self.options.append('--py2')

        if python_version:
            self.options.append('--python-version %s' % python_version)

        if strict_optional:
            self.options.append('--strict-optional')

        for list_option in MYPY_OPTIONS:
            for entry in options.get(list_option, []):
                self.options.append('--%s-%s' % (list_option, entry))

    def run(self, found_files):
        paths = [path for path in found_files.iter_module_paths()]
        paths.extend(self.options)
        result = self.checker.run(paths)
        report, _ = result[0], result[1:]
        messages = []

        for message in report.splitlines():
            iter_message = iter(message.split(':'))
            (path, line, char, err_type), err_msg = islice(iter_message, 4), list(message)
            location = Location(
                path=path,
                module=None,
                function=None,
                line=line,
                character=char,
                absolute_path=True
            )
            message = Message(
                source='mypy',
                code=err_type,
                location=location,
                message=''.join(err_msg).strip()
            )
            messages.append(message)

        return messages


from prospector.formatters.base import Formatter
from xml.dom.minidom import Document


class XunitFormatter(Formatter):

    """
    This formatter outputs messages in the Xunit xml format, which is used by several
    CI tools to parse output. This formatter is therefore a compatability shim between tools built
    to use Xunit and prospector itself.
    """

    def render(self, summary=True, messages=True, profile=False):
        xml_doc = Document()

        testsuite_el = xml_doc.createElement('testsuite')
        testsuite_el.setAttribute('errors', str(self.summary['message_count']))
        testsuite_el.setAttribute('failures', '0')
        testsuite_el.setAttribute('name', 'prospector-%s' % '-'.join(self.summary['tools']))
        testsuite_el.setAttribute('tests', str(self.summary['message_count']))
        testsuite_el.setAttribute('time', str(self.summary['time_taken']))
        xml_doc.appendChild(testsuite_el)

        prop_el = xml_doc.createElement('properties')
        testsuite_el.appendChild(prop_el)

        sysout_el = xml_doc.createElement('system-out')
        sysout_el.appendChild(xml_doc.createCDATASection(''))
        testsuite_el.appendChild(sysout_el)

        syserr_el = xml_doc.createElement('system-err')
        syserr_el.appendChild(xml_doc.createCDATASection(''))
        testsuite_el.appendChild(syserr_el)

        for message in sorted(self.messages):
            testcase_el = xml_doc.createElement('testcase')
            testcase_el.setAttribute('name', '%s-%s' % (message.location.path, message.location.line))

            failure_el = xml_doc.createElement('error')
            failure_el.setAttribute('message', message.message.strip())
            failure_el.setAttribute('type', '%s Error' % message.source)
            template = '%(path)s:%(line)s: [%(code)s(%(source)s), %(function)s] %(message)s'
            cdata = template % {
                'path': message.location.path,
                'line': message.location.line,
                'source': message.source,
                'code': message.code,
                'function': message.location.function,
                'message': message.message.strip()
            }
            failure_el.appendChild(xml_doc.createCDATASection(cdata))

            testcase_el.appendChild(failure_el)

            testsuite_el.appendChild(testcase_el)

        return xml_doc.toprettyxml()



# -*- coding: utf-8 -*-
from __future__ import absolute_import

from prospector.encoding import read_py_file, CouldNotHandleEncoding
from pydocstyle.checker import ConventionChecker as PEP257Checker, AllError
from prospector.message import Location, Message, make_tool_error_message
from prospector.tools.base import ToolBase


__all__ = (
    'Pep257Tool',
)


class Pep257Tool(ToolBase):
    def __init__(self, *args, **kwargs):
        super(Pep257Tool, self).__init__(*args, **kwargs)
        self._code_files = []
        self.ignore_codes = ()

    def configure(self, prospector_config, found_files):
        self.ignore_codes = prospector_config.get_disabled_messages('pep257')

    def run(self, found_files):
        messages = []

        checker = PEP257Checker()

        for code_file in found_files.iter_module_paths():
            try:
                for error in checker.check_source(
                    read_py_file(code_file),
                    code_file,
                    None
                ):

                    location = Location(
                        path=code_file,
                        module=None,
                        function='',
                        line=error.line,
                        character=0,
                        absolute_path=True,
                    )
                    message = Message(
                        source='pep257',
                        code=error.code,
                        location=location,
                        message=error.message.partition(':')[2].strip(),
                    )
                    messages.append(message)
            except CouldNotHandleEncoding as err:
                messages.append(make_tool_error_message(
                    code_file, 'pep257', 'D000',
                    message='Could not handle the encoding of this file: %s' % err.encoding
                ))
                continue
            except AllError as exc:
                # pep257's Parser.parse_all method raises AllError when an
                # attempt to analyze the __all__ definition has failed.  This
                # occurs when __all__ is too complex to be parsed.
                messages.append(make_tool_error_message(
                    code_file, 'pep257', 'D000',
                    line=1, character=0,
                    message=exc.args[0]
                ))
                continue

        return self.filter_messages(messages)

    def filter_messages(self, messages):
        return [
            message
            for message in messages
            if message.code not in self.ignore_codes
        ]



import os
import re
from prospector.formatters.base import Formatter


class VSCodeFormatter(Formatter):

    """
    This formatter outputs messages in the same way as vscode prospector linter expects.
    """

    def render(self, summary=True, messages=True, profile=False):
        # this formatter will always ignore the summary and profile
        cur_loc = None
        output = []

        for message in sorted(self.messages):

            if cur_loc != message.location.path:
                cur_loc = message.location.path
                module_name = message.location.path.replace(os.path.sep, '.')
                module_name = re.sub(r'(\.__init__)?\.py$', '', module_name)

                header = '************* Module %s' % module_name
                output.append(header)

            template = '%(line)s,%(character)s,%(code)s,%(code)s:%(source)s %(message)s'
            output.append(template % {
                'line': message.location.line,
                'character': message.location.character,
                'source': message.source,
                'code': message.code,
                'message': message.message.strip()
            })

        return '\n'.join(output)



# -*- coding: utf-8 -*-
from prospector.exceptions import FatalProspectorException
from prospector.tools.base import ToolBase
from prospector.tools.dodgy import DodgyTool
from prospector.tools.pep8 import Pep8Tool
from prospector.tools.pyflakes import PyFlakesTool
from prospector.tools.pylint import PylintTool
from prospector.tools.mccabe import McCabeTool
from prospector.tools.pep257 import Pep257Tool
from prospector.tools.profile_validator import ProfileValidationTool


def _tool_not_available(name, install_option_name):
    class NotAvailableTool(ToolBase):
        def run(self, _):
            raise FatalProspectorException("\nCannot run tool %s as support was not installed.\n"
                                           "Please install by running 'pip install prospector[%s]'\n\n"
                                           % (name, install_option_name))

    return NotAvailableTool


def _optional_tool(name, package_name=None, tool_class_name=None, install_option_name=None):
    package_name = 'prospector.tools.%s' % (package_name or name)
    tool_class_name = tool_class_name or '%sTool' % name.title()
    install_option_name = install_option_name or ('with_%s' % name)

    try:
        tool_package = __import__(package_name, fromlist=[tool_class_name])
    except ImportError:
        tool_class = _tool_not_available(name, install_option_name)
    else:
        tool_class = getattr(tool_package, tool_class_name)

    return tool_class


TOOLS = {
    'dodgy': DodgyTool,
    'mccabe': McCabeTool,
    'pyflakes': PyFlakesTool,
    'pep8': Pep8Tool,
    'pylint': PylintTool,
    'pep257': Pep257Tool,
    'profile-validator': ProfileValidationTool,
    'frosted': _optional_tool('frosted'),
    'vulture': _optional_tool('vulture'),
    'pyroma': _optional_tool('pyroma'),
    'mypy': _optional_tool('mypy'),
}


DEFAULT_TOOLS = (
    'dodgy',
    'mccabe',
    'pyflakes',
    'pep8',
    'pylint',
    'pep257',
    'profile-validator'
)



# -*- coding: utf-8 -*-
import sys
import tokenize


class CouldNotHandleEncoding(Exception):
    def __init__(self, path, cause):
        self.path = path
        self.cause = cause


def read_py_file(filepath):
    if sys.version_info < (3, ):
        return open(filepath, 'rU').read()
    else:
        # see https://docs.python.org/3/library/tokenize.html#tokenize.detect_encoding
        # first just see if the file is properly encoded
        try:
            with open(filepath, 'rb') as f:
                tokenize.detect_encoding(f.readline)
        except SyntaxError as err:
            # this warning is issued:
            #   (1) in badly authored files (contains non-utf8 in a comment line)
            #   (2) a coding is specified, but wrong and
            #   (3) no coding is specified, and the default
            #       'utf8' fails to decode.
            #   (4) the encoding specified by a pep263 declaration did not match
            #       with the encoding detected by inspecting the BOM
            raise CouldNotHandleEncoding(filepath, err)

        try:
            return tokenize.open(filepath).read()
            # this warning is issued:
            #   (1) if uft-8 is specified, but latin1 is used with something like \x0e9 appearing
            #       (see http://stackoverflow.com/a/5552623)
        except UnicodeDecodeError as err:
            raise CouldNotHandleEncoding(filepath, err)
